# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/profiler.ipynb.

# %% auto 0
__all__ = ['time_profiler']

# %% ../nbs/profiler.ipynb 2
import cProfile
import pstats
from io import StringIO

# %% ../nbs/profiler.ipynb 3
def time_profiler(filename="profiling_results.txt"):
    # The outer function that allows customization of the filename
    def decorator(func):
        # The middle function which receives the function to be wrapped
        def wrapper(*args, **kwargs):
            # The inner function that actually runs the profiling
            profiler = cProfile.Profile()
            profiler.enable()  # Start profiling
            result = func(*args, **kwargs)
            profiler.disable()  # End profiling
            # Create a StringIO stream to capture profiling results
            s = StringIO()
            ps = pstats.Stats(profiler, stream=s).sort_stats("cumulative")
            ps.print_stats()
            # Write the profiling results to the specified file
            with open(filename, "w") as f:
                f.write(s.getvalue())
            return result

        return wrapper

    return decorator



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/misc.ipynb.

# %% auto 0
__all__ = ['Timer', 'track2', 'summarize_input', 'timeit', 'io', 'tryy', 'cast_inputs']

# %% ../nbs/misc.ipynb 2
import time
from .logger import Debug, Warn, debug_mode, Info, Trace, trace_mode
from .markup2 import AD
from functools import wraps
from typing import get_type_hints
from fastcore.basics import ifnone
from fastcore.foundation import L

# %% ../nbs/misc.ipynb 3
class Timer:
    def __init__(self, N, smooth=True, mode=1):
        "print elapsed time every iteration and print out remaining time"
        "assumes this timer is called exactly N times or less"
        self.tok = self.start = time.time()
        self.N = N
        assert N is None or N >= 0, "N should be None or a non-negative integer"
        self.ix = 0
        self.smooth = smooth
        self.mode = mode
        # 0 = instant-speed, i.e., time remaining is a funciton of only the last iteration
        # Useful when you know that each loop takes unequal time (increasing/decreasing speeds)
        # 1 = average, i.e., time remaining is a function of average of all iterations
        # Usefule when you know on average each loop or a group of loops take around the same time

    def __call__(self, *, ix=None, info=None):
        ix = self.ix if ix is None else ix
        info = "" if info is None else f"{info}\t"
        tik = time.time()
        elapsed = tik - self.start

        _total = self.N if self.N is not None else "*"

        if self.mode == 0:
            ielapsed = tik - self.tok
            ispeed = ielapsed
            if self.N is not None:
                iremaining = (self.N - (ix + 1)) * ispeed
            else:
                iremaining = -1

            iunit = "s/iter"
            if ispeed < 1:
                ispeed = 1 / ispeed
                iunit = "iters/s"

            iestimate = iremaining + elapsed
            _remaining = f"- {iremaining:.2f}s remaining " if self.N is not None else ""

            _info = f"{info}{ix+1}/{_total} ({elapsed:.2f}s {_remaining}- {ispeed:.2f} {iunit})"

        else:
            speed = elapsed / (ix + 1)
            if self.N is not None:
                remaining = (self.N - (ix + 1)) * speed
            else:
                remaining = -1

            unit = "s/iter"
            if speed < 1:
                speed = 1 / speed
                unit = "iters/s"
            estimate = remaining + elapsed
            _remaining = f" - {remaining:.2f}s remaining " if self.N is not None else ""
            _info = f"{info}{ix+1}/{_total} ({elapsed:.2f}s {_remaining}- {speed:.2f} {unit})"

        print(
            _info + " " * 10,
            end="\r",
        )
        self.ix += 1
        self.tok = tik


def track2(iterable, *, total=None, info_prefix=None):
    info_prefix = ifnone(info_prefix, "")
    try:
        total = ifnone(total, len(iterable))
    except Exception as e:
        Warn(f"Unable to get length of iterable: {e}")
    timer = Timer(total)
    for item in iterable:
        info = yield item
        _info = (
            f"{info_prefix} {info}"
            if info and info_prefix
            else info or info_prefix or None
        )
        timer(info=_info)
        if info is not None:
            yield  # Just to ensure the send operation stops


# %% ../nbs/misc.ipynb 10
def summarize_input(args, kwargs, outputs=None):
    o = AD(args, kwargs)
    if outputs is not None:
        o.outputs = outputs
    return o.summary()


def timeit(func):
    def inner(*args, **kwargs):
        s = time.time()
        o = func(*args, **kwargs)
        Info(f"{time.time() - s:.2f} seconds to execute `{func.__name__}`")
        return o

    return inner


def io(func=None, *, level="debug"):
    logfuncs = {
        "debug": lambda i: Debug(i, depth=2),
        "info": lambda i: Info(i, depth=2),
        "trace": lambda i: Trace(i, depth=2),
    }
    try:
        logfunc = logfuncs[level.lower()]
    except KeyError:
        raise ValueError(f"level should be one of {list(logfuncs.keys())}")

    def decorator(func):
        def inner(*args, **kwargs):
            if level == "trace":
                try:
                    from pysnooper import snoop
                except ModuleNotFoundError:
                    raise ModuleNotFoundError(
                        "`pip install pysnooper` if you want to trace the function line by line"
                    )
                _func = snoop()(func)
            else:
                _func = func
            s = time.time()
            o = _func(*args, **kwargs)
            info = f"""
{time.time() - s:.2f} seconds to execute `{func.__name__}`
{summarize_input(args=args, kwargs=kwargs, outputs=o)}
            """
            logfunc(info)
            return o

        return inner

    if func is None:
        o = decorator
    else:
        o = decorator(func)
    return o

# %% ../nbs/misc.ipynb 18
def tryy(
    func=None,
    *,
    output_to_return_on_fail=None,
    silence_errors=False,
    print_traceback=False,
    store_errors: bool = True,
):
    def decorator(f):
        if isinstance(store_errors, bool) and store_errors:
            error_store = []
        elif isinstance(store_errors, (list, type([]))):  # Avoid `L` unless defined
            error_store = store_errors

        @wraps(f)  # Preserve the original function's metadata
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                if not silence_errors:
                    if not print_traceback:
                        tb = f"{type(e).__name__}: {str(e)}"
                        Warn(f"Error for `{f.__name__}`: {tb}")
                    else:
                        import traceback

                        tb = traceback.format_exc()
                        Warn(f"Error for `{f.__name__}`:\n{tb}")
                else:
                    tb = None
                if store_errors is not None:
                    error_store.append(
                        {
                            "func": f.__name__,
                            "args": args,
                            "kwargs": kwargs,
                            "tb": tb,
                            "err_type": type(e).__name__,
                        }
                    )
                if callable(output_to_return_on_fail):
                    return output_to_return_on_fail(*args, **kwargs)
                return output_to_return_on_fail

        # Attach additional attributes
        wrapper.F = f
        wrapper.error_store = error_store

        def errors():
            import pandas as pd

            return pd.DataFrame(error_store)

        wrapper.errors = errors
        wrapper.error_summary = errors  # backward compatibility
        return wrapper

    if callable(func):
        return decorator(func)
    return decorator

# %% ../nbs/misc.ipynb 40
def cast_inputs(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Get type hints of the function
        type_hints = get_type_hints(func)

        # Get the function arguments as a dictionary
        from inspect import signature

        bound_args = signature(func).bind(*args, **kwargs)
        bound_args.apply_defaults()

        # Cast arguments to the specified type if possible
        for arg_name, arg_value in bound_args.arguments.items():
            if arg_name in type_hints:
                target_type = type_hints[arg_name]
                try:
                    bound_args.arguments[arg_name] = target_type(arg_value)
                except (ValueError, TypeError):
                    raise TypeError(f"Cannot cast `{arg_name}` ({arg_value}) to {target_type}")

        # Call the function with the cast arguments
        return func(*bound_args.args, **bound_args.kwargs)

    return wrapper




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/s3_loader2.ipynb.

# %% auto 0
__all__ = ['S3FileHandler']

# %% ../nbs/s3_loader2.ipynb 1
import boto3
import os
import datetime
from datetime import tzinfo
from dateutil.tz import tzutc
from . import stem, fname

# %% ../nbs/s3_loader2.ipynb 3
class S3FileHandler:
    def __init__(self, aws_access_key, aws_secret_access_key):
        self.s3_client = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

    def list_s3_buckets(self):
        """
        Lists all the s3 buckets in s3
        """
        try:
            # Call S3 to list current buckets
            response = self.s3_client.list_buckets()
            buckets = [bucket["Name"] for bucket in response["Buckets"]]
            return buckets
        except Exception as e:
            print(e)

    def list_s3_objects(self, bucket_name, key=""):
        """
        List all files in an S3 bucket or within a specific prefix.

        :param bucket_name: str. Name of the S3 bucket.
        :param key: str or None. Specific prefix to list files from, defaults to None.
        """
        try:
            # Initialize a paginator for listing objects
            paginator = self.s3_client.get_paginator("list_objects_v2")
            # Use the paginator to fetch all objects in the specified bucket and prefix if provided
            files = dict()
            for page in paginator.paginate(Bucket=bucket_name, Prefix=key):
                # Access the 'Contents' from the page, which lists the objects
                if "Contents" in page:
                    for obj in page["Contents"]:
                        files[obj["Key"]] = obj["Size"]
                        # print(f"{obj['Key']} ({obj['Size']} bytes)")
            return files
        except Exception as e:
            print(f"An error occurred: {e}")

    def download_s3_folder(self, bucket_name, local_dir, prefix="", verbose=0):
        """
        Download all files from an S3 bucket prefix to a local directory.

        :param bucket_name: str. Name of the S3 bucket.
        :param local_dir: str. Local directory to which files will be downloaded.
        :param prefix: str or None. Prefix path of the folder in the bucket. If None, the whole bucket is downloaded.
        """
        if not prefix.endswith("/"):
            prefix = prefix + "/"
            print(prefix)
        # Ensure local directory exists
        if prefix == "":
            local_dir = os.path.join(local_dir, bucket_name)
        else:
            local_dir = os.path.join(local_dir, stem(prefix))
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)

        # List objects within the specified prefix
        paginator = self.s3_client.get_paginator("list_objects_v2")
        for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
            for obj in page.get("Contents", []):
                key = obj["Key"]
                if not key.endswith("/"):  # skip directories/folders
                    # Define file path locally in same structure
                    local_file_path = os.path.join(local_dir, key[len(prefix) :])
                    local_file_dir = os.path.dirname(local_file_path)

                    # Ensure local file directory exists
                    if not os.path.exists(local_file_dir):
                        os.makedirs(local_file_dir)

                    # Download the file
                    self.s3_client.download_file(bucket_name, key, local_file_path)
                    if verbose:
                        print(f"Downloaded {key} to {local_file_path}")

    def download_s3_file(self, bucket_name, key, local_dir, metadata=False, verbose=0):
        """
        Download a specific file from an S3 bucket and optionally return its metadata.

        :param bucket_name: str. Name of the S3 bucket.
        :param key: str. The key of the file in the S3 bucket.
        :param local_dir: str. Local directory to which the file will be downloaded.
        :param metadata: bool. If True, return the file's metadata; otherwise, return None.
        :param verbose: bool.
        :return: dict or None. Returns metadata of the file if metadata is True, otherwise None.
        """
        # Define the local file path
        local_file_path = os.path.join(local_dir, os.path.basename(key))

        # Ensure the local directory exists
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)

        # Download the file
        self.s3_client.download_file(bucket_name, key, local_file_path)
        if verbose:
            print(f"Downloaded {key} to {local_file_path}")

        # Optionally retrieve and return metadata
        if metadata:
            response = self.s3_client.head_object(Bucket=bucket_name, Key=key)
            return response  # Return the metadata dictionary
        return None

    def upload_file_to_s3(self, bucket_name, localfile_path, s3_key, metadata=None):
        """
        Upload a file to an S3 bucket with optional metadata.

        :param bucket_name: str. Name of the S3 bucket.
        :param localfile_path: str. Local path to the file to be uploaded.
        :param s3_key: str. S3 key (path within the bucket) where the file will be stored with file name included.
        :param metadata: dict or None. Optional metadata for the file. Defaults to None.
        """
        try:
            # Setup the file upload options
            extra_args = {}
            if metadata:
                extra_args["Metadata"] = metadata

            # Perform the file upload
            with open(localfile_path, "rb") as file_data:
                self.s3_client.upload_fileobj(
                    Fileobj=file_data,
                    Bucket=bucket_name,
                    Key=s3_key,
                    ExtraArgs=extra_args,
                )
            print(f"File uploaded successfully to {bucket_name}/{s3_key}")
        except Exception as e:
            print(f"Failed to upload file: {e}")

    def upload_folder_to_s3(
        self, bucket_name, local_folder_path, s3_prefix="", metadata=None, verbose=0
    ):
        """
        Upload all files in a local folder to an S3 bucket with optional metadata.

        :param bucket_name: str. Name of the S3 bucket.
        :param local_folder_path: str. Local path to the folder to be uploaded.
        :param s3_prefix: str. S3 prefix (folder path within the bucket) where the files will be stored.
                           Defaults to the root of the bucket.
        :param metadata: dict or None. Optional metadata for the files. Defaults to None.
        """
        try:
            # Ensure the local_folder_path ends with a slash to properly preserve folder structure
            local_folder_path = os.path.normpath(local_folder_path)

            # Iterate over all files in the folder and its subfolders
            for root, dirs, files in os.walk(local_folder_path):
                for filename in files:
                    local_file_path = os.path.join(root, filename)

                    # Generate the corresponding S3 key (prefix + relative file path)
                    relative_path = os.path.relpath(
                        local_file_path, os.path.dirname(local_folder_path)
                    )
                    s3_key = os.path.join(s3_prefix, relative_path).replace(
                        "\\", "/"
                    )  # Replace Windows path separators

                    extra_args = {}
                    if metadata:
                        extra_args["Metadata"] = metadata

                    with open(local_file_path, "rb") as file_data:
                        self.s3_client.upload_fileobj(
                            Fileobj=file_data,
                            Bucket=bucket_name,
                            Key=s3_key,
                            ExtraArgs=extra_args,
                        )
                    if verbose:
                        print(f"Uploaded {s3_key} to {bucket_name}/{s3_key}")

        except Exception as e:
            print(f"Failed to upload folder: {e}")

    def inmemory_download_s3(bucket_name, key):
        """
        Downloads a file from an Amazon S3 bucket and loads it directly into a pandas DataFrame.
        The function automatically detects the file format based on its extension.

        Parameters:
        key (str): The S3 object key of the file to download.
        bucket (str, optional): The name of the S3 bucket. Defaults to AWS_BUCKET from .env if not provided.
        """
        response = self.s3_client.get_object(Bucket=bucket_name, Key=key)
        file_content = response["Body"].read()
        return file_content



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/imgaug_loader.ipynb.

# %% auto 0
__all__ = ['do', 'bw', 'rotate', 'pad', 'get_size', 'rescale', 'crop', 'imgaugbbs2bbs', 'bbs2imgaugbbs']

# %% ../nbs/imgaug_loader.ipynb 2
import imgaug.augmenters as iaa
from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage
from .loader import BB, PIL, bbs2df, df2bbs, np, pd, Image
from torch_snippets.bb_utils import (
    split_bb_to_xyXY,
    combine_xyXY_to_bb,
    to_relative,
    to_absolute,
)

# %% ../nbs/imgaug_loader.ipynb 3
def do(img, bbs=None, aug=None, cval=255):
    """
    Apply image augmentation to the input image and bounding boxes.

    Args:
        img (numpy.ndarray or PIL.Image.Image): The input image.
        bbs (pandas.DataFrame or None): The bounding boxes associated with the image.
        aug (imgaug.augmenters.Augmenter or None): The image augmentation object.
        cval (int): The constant value used for padding.

    Returns:
        tuple or PIL.Image.Image: If `bbs` is None, returns the augmented image.
        Otherwise, returns a tuple containing the augmented image and the augmented bounding boxes.
    """
    if isinstance(img, PIL.Image.Image):
        _Image = True
        img = np.array(img)
    else:
        _Image = False
    no_bbs = False
    if bbs is None:
        no_bbs = True
        bbs = []
    H, W = img.shape[:2]
    if isinstance(bbs, pd.DataFrame):
        _df = bbs.copy()
        _separate = True if "x" in _df.columns else False
        if not _separate:
            _df = split_bb_to_xyXY(_df)
        _relative = True if _df["x"].max() < 1 else False
        if _relative:
            _df = to_absolute(_df, H, W)
        bbs = df2bbs(_df)
        remaining_columns = [c for c in _df.columns if c not in "xyXY"]
        __df = _df[remaining_columns]
        _data_frame = True
    else:
        _data_frame = False

    bbs = bbs2imgaugbbs(bbs, img)
    img, bbs = aug(images=[img], bounding_boxes=[bbs])
    img, bbs = (img[0], imgaugbbs2bbs(bbs))
    H, W = img.shape[:2]

    if _Image:
        img = Image.fromarray(img)
    if _data_frame:
        _df = bbs2df(bbs)
        __df[[*"xyXY"]] = _df.values
        if _relative:
            __df = to_relative(__df, H, W)
        if not _separate:
            __df = combine_xyXY_to_bb(__df)
        bbs = __df
    if no_bbs:
        return img
    return img, bbs


def bw(img, bbs):
    """
    Applies grayscale augmentation to the input image.

    Args:
        img (numpy.ndarray): The input image.
        bbs (list): List of bounding boxes associated with the image.

    Returns:
        numpy.ndarray: The augmented image.

    """
    aug = iaa.Grayscale()
    return do(img, bbs, aug)


def rotate(img, bbs=None, angle=None, cval=255):
    """
    Rotate the input image and bounding boxes (if provided) by a given angle.

    Args:
        img (numpy.ndarray): The input image.
        bbs (list, optional): List of bounding boxes. Defaults to None.
        angle (float, optional): The angle of rotation in degrees. Defaults to None.
        cval (int, optional): The constant value used to fill the empty space after rotation. Defaults to 255.

    Returns:
        numpy.ndarray: The rotated image.
    """
    aug = iaa.Rotate(angle, cval=cval, fit_output=True)
    return do(img, bbs=bbs, aug=aug)


def pad(img, bbs, sz=None, deltas=None, cval=0):
    """
    Pad an image and its bounding boxes.

    Args:
        img (numpy.ndarray or PIL.Image.Image): The input image.
        bbs (list): List of bounding boxes.
        sz (tuple, optional): The desired size of the output image. If provided, the image will be padded to this size. Defaults to None.
        deltas (tuple, optional): The amount of padding to be applied on each side of the image. If provided, sz will be ignored. Defaults to None.
        cval (int, optional): The value used for padding. Defaults to 0.

    Returns:
        numpy.ndarray: The padded image.
    """
    if isinstance(img, np.ndarray):
        h, w = img.shape[:2]
    else:
        w, h = img.size
    if sz:
        H, W = sz
        deltas = (H - h) // 2, (W - w) // 2, (H - h) // 2, (W - w) // 2

    aug = iaa.Pad(deltas, pad_cval=cval)
    return do(img, bbs, aug)


def get_size(sz, h, w):
    """
    Calculate the target size (height and width) based on the input size and resize parameters.

    Args:
        sz (tuple, list, float, int): The resize parameters. It can be one of the following:
            - (tuple, list): A tuple or list containing a signal and target size (H, W).
                             The signal can be either 'at-least' or 'at-most'.
                             The target size represents the desired size of the image.
            - float: A float value representing the fraction of the input size.
            - int: An integer value representing the target size.
            - tuple: A tuple containing the target size (H, W).
                     The target size can be -1 to maintain the aspect ratio of the input size.
            - float: A float value representing the target size as a fraction of the input size.

        h (int): The height of the input size.
        w (int): The width of the input size.

    Returns:
        tuple: A tuple containing the target size (H, W).

    Raises:
        AssertionError: If the resize type is not 'at-least' or 'at-most'.

    """
    if isinstance(sz, (tuple, list)) and isinstance(sz[0], str):
        signal, (H, W) = sz
        assert signal in "at-least,at-most".split(
            ","
        ), "Resize type must be one of `at-least` or `at-most`"
        if signal == "at-least":
            f = max(H / h, W / w)
        if signal == "at-most":
            f = min(H / h, W / w)
        H, W = [i * f for i in [h, w]]
    elif isinstance(sz, float):
        frac = sz
        H, W = [i * frac for i in [h, w]]
    elif isinstance(sz, int):
        H, W = sz, sz
    elif isinstance(sz, tuple):
        H, W = sz
        if H == -1:
            _, W = sz
            f = W / w
            H = f * h
        elif W == -1:
            H, _ = sz
            f = H / h
            W = f * w
        elif isinstance(H, float):
            H = H * h
        elif isinstance(W, float):
            W = W * h
    H, W = int(H), int(W)
    return H, W


def rescale(im, bbs, sz):
    """
    Rescales the input image and bounding boxes to the specified size.

    Args:
        im (PIL.Image.Image or numpy.ndarray): The input image.
        bbs (list): List of bounding boxes.
        sz (tuple): The target size (height, width) to resize the image.

    Returns:
        PIL.Image.Image: The resized image.
        list: The resized bounding boxes.

    """
    if isinstance(im, PIL.Image.Image):
        to_pil = True
        im = np.array(im)
    else:
        to_pil = False
    h, w = im.shape[:2]
    H, W = get_size(sz, h, w)
    aug = iaa.Resize({"height": H, "width": W})
    im, bbs = do(im, bbs, aug)
    if to_pil:
        im = PIL.Image.fromarray(im)
    return im, bbs


def crop(img, bbs, deltas):
    """
    Crop the image and bounding boxes using the specified deltas.

    Args:
        img (numpy.ndarray): The input image.
        bbs (list): List of bounding boxes.
        deltas (tuple or list): The crop deltas in the form of (top, right, bottom, left).

    Returns:
        numpy.ndarray: The cropped image.
        list: The cropped bounding boxes.
    """
    aug = iaa.Crop(deltas)
    return do(img, bbs, aug)


def imgaugbbs2bbs(bbs):
    """
    Converts a list of imgaug bounding boxes to a list of custom BB objects.

    Args:
        bbs (list): A list of imgaug bounding boxes.

    Returns:
        list: A list of custom BB objects.

    """
    if bbs is None:
        return None
    return [
        BB([int(i) for i in (bb.x1, bb.y1, bb.x2, bb.y2)])
        for bb in bbs[0].bounding_boxes
    ]


def bbs2imgaugbbs(bbs, img):
    """
    Convert a list of bounding boxes to an imgaug BoundingBoxesOnImage object.

    Args:
        bbs (list): List of bounding boxes in the format [(x1, y1, x2, y2), ...].
        img (numpy.ndarray): Input image.

    Returns:
        imgaug.BoundingBoxesOnImage: BoundingBoxesOnImage object representing the bounding boxes on the image.
    """
    if bbs is None:
        return None
    return BoundingBoxesOnImage(
        [BoundingBox(x1=x, y1=y, x2=X, y2=Y) for x, y, X, Y in bbs], shape=img.shape
    )



__version__ = "0.555"
from .logger import *
from .loader import *
from .paths import *
from .markup import *
from .markdown import *

# from .inspector import *
from .load_defaults import *
from .pdf_loader import PDF
from .markup2 import AD
from .registry import *
from .ipython import *
from .decorators import *
from .misc import *
from .dates import *
from .s3_loader import *
from .zen import *

# Optional lazy-load safe placeholders for linters
torch = nn = F = th = T = transforms = torchvision = np = Dataset = DataLoader = (
    optim
) = Report = Reshape = Permute = device = save_torch_model_weights_from = (
    load_torch_model_weights_to
) = detach = cat_with_padding = None

def init_torch():
    try:
        from .torch_loader import (
            torch,
            th,
            torchvision,
            T,
            transforms,
            nn,
            np,
            F,
            Dataset,
            DataLoader,
            optim,
            Report,
            Reshape,
            Permute,
            device,
            save_torch_model_weights_from,
            load_torch_model_weights_to,
            detach,
            cat_with_padding,
        )

        import lovely_tensors

        lovely_tensors.monkey_patch()
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Unable to install torch dependencies. Please install them using `pip install torch-snippets[torch]`"
        )

    import builtins

    # Define a dictionary mapping names to imported objects
    modules = {
        "torch": torch,
        "th": th,
        "torchvision": torchvision,
        "T": T,
        "transforms": transforms,
        "nn": nn,
        "np": np,
        "F": F,
        "Dataset": Dataset,
        "DataLoader": DataLoader,
        "optim": optim,
        "Report": Report,
        "Reshape": Reshape,
        "Permute": Permute,
        "device": device,
        "save_torch_model_weights_from": save_torch_model_weights_from,
        "load_torch_model_weights_to": load_torch_model_weights_to,
        "detach": detach,
        "cat_with_padding": cat_with_padding,
    }

    # Use a loop to set attributes in builtins
    for name, module in modules.items():
        setattr(builtins, name, module)



from markitdown import MarkItDown
import os
from typer import Option
from .logger import Info
from .paths import P
from .cli import cli
from .misc import track2


def convert_to_markdown(input_file, output_dir, md_parser, root_path=None):
    """Converts a single file to markdown"""
    try:
        if root_path is None:
            output_file_path = output_dir / f"{input_file.stem}.md"
        else:
            relative_path = input_file.relative_to(root_path)
            output_file_path = output_dir / relative_path.with_suffix(f"{input_file.suffix}.md")

        if os.path.exists(output_file_path):
            Info(f"File {output_file_path} already exists, skipping.")
            return None

        result = md_parser.convert(str(input_file))
        output_file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(output_file_path, "w") as f:
            f.write(result.text_content)
        return f"Converted {input_file} to {output_file_path}"
    except Exception as e:
        return f"Could not convert {input_file}. Reason: {e}"


@cli.command('mkd')
def create_markdown(input_path: str, output_path: str = Option(None, "-o", "--output")):
    """
    Recursively creates markdown files from a source folder or a single file to a destination folder.
    """
    input_path = P(input_path).resolve()

    if output_path is None:
        output_path = input_path

    output_path = P(output_path)

    if not output_path.exists():
        output_path.mkdir(parents=True)

    md = MarkItDown()

    if input_path.is_file():
        result = convert_to_markdown(input_path, output_path, md)
        print(result)
        return

    for root, _, files in os.walk(input_path):
        tracker = track2(files)
        for file in tracker:
            file_path = P(root) / file
            result = convert_to_markdown(file_path, output_path, md, root_path=input_path)
            if result is None:
                continue
            if "Could not convert" in result:
                print(result)
            else:
                tracker.send(result)




from .cli import cli
from .paths import P, parent, makedir, fname
from .load_defaults import exists
from .logger import Info, notify_waiting
from .misc import tryy


class S3Copier:
    def __init__(self):
        from s3fs import S3FileSystem

        self.s3 = S3FileSystem()

    @tryy
    def _download(self, s3_f, local_file):
        recursive = "." in s3_f
        with notify_waiting(f"Downloading\nSOURCE: {s3_f}\nDESTIN: {local_file}"):
            self.s3.download(
                str(s3_f), str(local_file), recursive=recursive
            )  # Download the object

    def download_s3_folder(
        self, s3_location, local_location, completeness_file, exclusion_list=None
    ):
        completeness_file = P(completeness_file)
        local_location = P(local_location)
        makedir(parent(local_location))
        if isinstance(exclusion_list, str):
            exclusion_list = [e.strip() for e in exclusion_list.split(",")]

        if not exists(completeness_file):
            objects = self.s3.listdir(s3_location)  # List objects in the S3 folder
            # objects = {_s: set(stems([k['Key'] for k in objects]))}
            objects = {fname(k["Key"]): k["Key"] for k in objects}
            for obj in objects:
                if exclusion_list and any(
                    excluded in obj for excluded in exclusion_list
                ):
                    continue  # Skip if the object matches any exclusion pattern
                s3_f = f"s3://{objects[obj]}"
                local_file = local_location / obj  # Define the local file path
                self._download(s3_f, local_file)
            completeness_file.touch()

    def download_s3_file(self, s3_location, local_location):
        makedir(parent(local_location))
        self.s3.download(s3_location, local_location)

    def upload_s3_folder(self, local_location, s3_location):
        makedir(parent(local_location))
        self.s3.upload(local_location, s3_location, recursive=True)

    def upload_s3_file(self, local_location, s3_location):
        makedir(parent(local_location))
        self.s3.upload(local_location, s3_location)


@cli.command()
def download_s3_folder(
    s3_location, local_location, completeness_file, exclusion_list=None
):
    S3Copier().download_s3_folder(
        s3_location, local_location, completeness_file, exclusion_list
    )


@cli.command()
def download_s3_file(s3_location, local_location):
    S3Copier().download_s3_file(s3_location, local_location)


@cli.command()
def upload_s3_folder(local_location, s3_location):
    S3Copier().upload_s3_folder(local_location, s3_location)


@cli.command()
def upload_s3_file(local_location, s3_location):
    S3Copier().upload_s3_file(local_location, s3_location)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/markups.ipynb.

# %% auto 0
__all__ = [
    "AD",
    "Config",
    "isnamedtupleinstance",
    "unpack",
    "AttrDict",
    "pretty_json",
    "read_json",
    "write_json",
    "write_jsonl",
    "read_jsonl",
    "read_yaml",
    "write_yaml",
    "read_xml",
    "write_xml",
    "basic_ad_repr",
]

import hashlib
from copy import deepcopy

# %% ../nbs/markups.ipynb 2
import json
import os
from collections.abc import Mapping
from json import JSONEncoder
from typing import Union, List
from dataclasses import is_dataclass, asdict
import jsonlines
import xmltodict
import yaml
from deepdiff import DeepDiff

from .icecream import ic
from .loader import BB, L, np, pd
from .logger import *
from .paths import *
from .thinc_parser.parser import Config


# %% ../nbs/markups.ipynb 3
def _default(self, obj):
    import numpy as np
    from datetime import datetime, date

    if isinstance(obj, P):
        return str(obj)
    if isinstance(obj, AD):
        return obj.dict()
    if isinstance(obj, (set, L)):
        return list(obj)
    try:
        import torch

        if isinstance(obj, torch.Tensor):
            obj = obj.cpu().detach().numpy()
    except:
        ...
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    return getattr(obj.__class__, "__json__", _default.default)(obj)


_default.default = JSONEncoder().default
JSONEncoder.default = _default


def isnamedtupleinstance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i) == str for i in fields)


def unpack(obj):
    if isinstance(obj, dict):
        return {key: unpack(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [unpack(value) for value in obj]
    elif isnamedtupleinstance(obj):
        return {key: unpack(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(unpack(value) for value in obj)
    else:
        return obj


def hash_tensor(tensor, n=8):
    import torch

    assert isinstance(tensor, torch.Tensor)
    try:
        tensor_str = tensor.cpu().detach().numpy().tobytes()
    except:
        ...
    hash_obj = hashlib.sha256(tensor_str)
    return "ID:#" + hash_obj.hexdigest()[:n]


def hash_pandas_dataframe(input):
    try:
        from pandas.util import hash_pandas_object

        h = hash_pandas_object(input, index=True).values
        return "ID:#" + hashlib.sha256(h).hexdigest()[:6]
    except:
        return "ID:#<uncomputable>"


def is_attrdict_like(_input):
    o = isinstance(_input, (Mapping, AttrDict)) or all(
        [hasattr(_input, k) for k in ["keys", "values", "items"]]
    )
    return o


class AttrDict(object):
    """
    Utility class to interact with a dictionary as if it were an object. `AD` is an alias to this class

    FEATURES:
    0. Access and modify keys (including nested keys) as if they were object attributes, supporting tab-completion.
       Example: `self.key1.key2[0].key3`
    1. Keys and values are recursively converted to AttrDict instances.
    2. Pretty-print the dictionary using `print`.
    3. Convert the entire structure to a regular dictionary at any time using `self.to_dict() / self.dict()`.
    3. Recursively remove keys using `self.drop(key)` from a JSON object.
    4. Apply a function to all values at all levels using `map`.

    GOTCHAS:
    1. All integer keys are implicitly converted to strings due to the enforced `self.key` format.
    2. You can still use `self[int]`, but this internally converts the integer to a string.

    METHODS:
    - `items()`: Return the items of the AttrDict as key-value pairs.
    - `keys()`: Return the keys of the AttrDict.
    - `values()`: Return the values of the AttrDict.
    - `update(dict)`: Update the AttrDict with key-value pairs from another dictionary.
    - `get(key, default=None)`: Get the value associated with a key, with an optional default value.
    - `__iter__()`: Allow iteration over the keys of the AttrDict.
    - `__len__()`: Return the number of keys in the AttrDict.
    - `__repr__()`: Return a string representation of the AttrDict.
    - `__dir__()`: List the keys of the AttrDict as attributes.
    - `__contains__(key)`: Check if a key exists in the AttrDict, use 'a.b.c' notation to directly check for a nested attribute.
    - `__delitem__(key)`: Delete a key from the AttrDict.
    - `map(func)`: Apply a function to all values in the AttrDict.
    - `drop(key)`: Recursively remove a key and its values from the AttrDict.
    - `to_dict()`: Convert the AttrDict and its nested structure to a regular dictionary.
    - `pretty(print_with_logger=False, *args, **kwargs)`: Pretty-print the AttrDict as JSON.
    - `__eq__(other)`: Compare the AttrDict with another dictionary for equality.
    - `find_address(key, current_path="")`: Find and return all addresses (paths) of a given key in the AttrDict.
    - `summary(current_path='', summary_str='', depth=0, sep='\t')`: Generate a summary of the structure and values in the AttrDict.
    - `write_summary(to, **kwargs)`: Write the summary to a file or stream.
    - `fetch(addr)`: Retrieve a value at a specified address (path).

    PARAMETERS:
    - `data` (dict, optional): Initial data to populate the AttrDict.

    USAGE:
    - Create an AttrDict instance by providing an optional initial dictionary, and then access and manipulate its contents as if they were object attributes.

    EXAMPLE:
    ```python
    my_dict = {'name': 'John', 'age': 30, 'address': {'city': 'New York', 'zip': '10001'}}
    attr_dict = AttrDict(my_dict)
    print(attr_dict.name)  # Access values like attributes
    attr_dict.address.city = 'Los Angeles'  # Modify nested values
    ```
    """

    forbidden = set(":,'\"}{.")

    def __init__(self, *args, given_input_to_ad=None, **kwargs):
        given_input_to_ad = {} if given_input_to_ad is None else given_input_to_ad
        if len(args) == 1 and isinstance(args[0], (Mapping, AttrDict)):
            given_input_to_ad = args[0]
            args = {}
        else:
            _args = dict(ic.io(*args)) if len(args) > 0 else {}
            args = {}
            for k in reversed(list(_args.keys())):
                v = _args[k]
                if any(c in self.forbidden for c in k):
                    assert isinstance(
                        v, (dict, AttrDict)
                    ), f"Input `{v}` can't be a list"
                    given_input_to_ad = {**v, **given_input_to_ad}
                else:
                    args = {**{k: v}, **args}

        assert not any(
            k in list(kwargs.keys()) for k in args.keys()
        ), "Keys in args and kwargs overlap"
        given_input_to_ad = {**args, **given_input_to_ad, **kwargs}
        for name, value in given_input_to_ad.items():
            setattr(self, str(name), self._wrap(value))

    def __call__(self, **kwargs):
        o = deepcopy(self)
        for key in kwargs:
            o[key] = kwargs[key]

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __json__(self):
        return self.dict()

    def _wrap(self, value):
        if isinstance(value, (L, tuple, list, set, frozenset)):
            value = type(value)([self._wrap(v) for v in value])
            if isinstance(value, (list, L)):
                value = L(value)
            return value
        else:
            return (
                AttrDict(given_input_to_ad=value) if isinstance(value, dict) else value
            )

    __getitem__ = lambda self, x: (
        AttrDict({_x: self[_x] for _x in x})
        if isinstance(x, (list, L))
        else getattr(self, str(x))
    )
    __setitem__ = lambda self, k, v: setattr(self, str(k), self._wrap(v))

    def update(self, input):
        input = AD(input)
        for k, v in input.items():
            self[k] = v

    def get(self, key, default=None):
        return self.fetch(key) if key in self else default

    def set(self, key, value):
        if "." in key:
            _keys = key.split(".")
            leaf = _keys[-1]
        else:
            _keys = [key]
            leaf = key
        for _key in _keys:
            if _key not in self and key != leaf:
                self[_key] = AD()
            if _key == leaf:
                self[_key] = value
            self = self[_key]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def __repr__(self):
        return f"\n```↯ AttrDict ↯\n{self.summary()}\n```\n"

    def __dir__(self):
        return self.__dict__.keys()

    def __contains__(self, key):
        key = str(key)
        if "." not in key:
            return key in self.__dict__.keys()
        else:
            d = self
            for _k in key.split("."):
                try:
                    d = d[_k]
                except AttributeError:
                    return False
            return True

    def __delitem__(self, key):
        key = str(key)
        del self.__dict__[key]

    def map(self, func, _try=False):
        if _try:

            def tryfunc(func):
                def inner(x):
                    try:
                        o = func(x)
                    except:
                        o = x
                    return o

                return inner

            _func = tryfunc(func)

        else:
            _func = func

        d = deepcopy(self)
        for k in dir(d):
            v = d[k]
            if isinstance(v, AttrDict):
                v = v.map(_func)
            elif isinstance(v, (L, tuple, list, set, frozenset)):
                v = [
                    _v.map(_func) if isinstance(_v, AttrDict) else _func(_v) for _v in v
                ]
            else:
                v = _func(v)
            d[k] = v
        return d

    def trymap(self, func):
        return self.map(func, _try=True)

    def drop(self, key, inplace=True):
        if not inplace:
            d = deepcopy(self)
            d.drop(key)
            return d
        if key in self:
            del self[key]
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.drop(key)
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.drop(key) for _v in v if isinstance(_v, AttrDict)]

    def dict(self):
        d = {}
        for k in self.__dict__.keys():  # can't use dir here
            v = self[k]
            if isinstance(v, AttrDict):
                v = v.dict()
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = type(v)([_v.dict() if isinstance(_v, AttrDict) else _v for _v in v])
            d[k] = v
        return d

    dict = dict

    def pretty(self, print_with_logger=False, *args, **kwargs):
        pretty_json(self.dict(), print_with_logger=print_with_logger, *args, **kwargs)

    def __eq__(self, other):
        return AttrDict(given_input_to_ad=other).dict() == self.dict()

    def find_address(self, key, current_path=""):
        if isinstance(key, (list, L)):
            return L([a for _key in key for a in self.find_address(_key, current_path)])
        addresses = []
        for k in self.keys():
            if current_path:
                new_path = f"{current_path}.{k}"
            else:
                new_path = k

            if k == key:
                addresses.append(new_path)

            if isinstance(self[k], AttrDict):
                addresses.extend(self[k].find_address(key, new_path))

            elif isinstance(self[k], (L, tuple, list, set, frozenset)):
                for i, item in enumerate(self[k]):
                    if isinstance(item, AttrDict):
                        addresses.extend(item.find_address(key, f"{new_path}.{i}"))
        return addresses

    def summary(self, current_path="", depth=0, sep="  ", max_items=10):
        max_items = int(os.environ.get("AD_MAX_ITEMS", max_items))
        if os.environ.get("AD_SHOW_TAB_STOPS", False):
            sep = "⋮ "
        sep = os.environ.get("AD_SEP", sep)

        def format_path(path, key):
            return f"{path}.{key}" if path else key

        def format_item(key, item, path, depth, sep):
            import numpy as np
            import pandas as pd

            try:
                import torch
            except ModuleNotFoundError:

                class Torch:
                    Tensor = type(...)

                torch = Torch()

            if is_dataclass(item):
                _type = type(item).__name__
                item = AD(asdict(item))
                info = f"(🏷️ {_type}:dataclass)"
            else:
                info = None

            if isinstance(item, (pd.DataFrame,)):
                return f"{sep * depth}{key} - {type(item).__name__} - shape {item.shape} - columns {item.columns} - {hash_pandas_dataframe(item)}\n"
            if is_attrdict_like(item):
                try:
                    item = AttrDict(**item)
                    _info = "" if info is None else info
                    return f"{sep*depth}{key}{_info}\n" + item.summary(
                        path, depth + 1, sep
                    )
                except:
                    return ""
            elif isinstance(item, (list, tuple, set, frozenset, L)):
                return summarize_collection(key, item, path, depth + 1, sep)
            elif isinstance(item, (torch.Tensor, np.ndarray)):
                is_np = False
                if isinstance(item, np.ndarray):
                    is_np = True
                    try:
                        item = torch.tensor(item)
                    except:
                        item = torch.tensor(item.copy())
                is_np = "🔦" if not is_np else "np."
                return f"{sep * depth}{key} - {is_np}{item} - {hash_tensor(item)}\n"

            else:
                is_multiline = False
                ogitem = item
                item = str(item)
                if isinstance(item, (str, P)):
                    item = str(item)
                    is_multiline = "\n" in item
                    if not os.environ.get("AD_SHOW_FULL_STRINGS", False):
                        _sep = (
                            " ...\n...\n...\n...\n... " if is_multiline else "........."
                        )
                        if len(item) > 100:
                            item = item[:35] + _sep + item[-35:]
                    else:
                        _sep = ""
                    if is_multiline:
                        _item = item.split("\n")
                        _item = "\n".join([f"{sep*(depth+1)}{l}" for l in _item])
                        item = f"↓\n{sep*(depth+1)}```\n{_item}\n{sep*(depth+1)}```"
                multiline = (
                    ""
                    if not (is_multiline and isinstance(ogitem, str))
                    else "Multiline "
                )
                return f"{sep * depth}{key} - {item} (🏷️ {multiline}{type(ogitem).__name__})\n"

        def summarize_collection(key, collection, path, d, s):
            if isinstance(collection, (L, list)):
                info = "[]"
            elif isinstance(collection, tuple):
                info = "()"
            else:
                info = "{}"
            summary_str = f"{s * (d - 1)}{key}{info}\n"
            for i, item in enumerate(collection):
                item_path = format_path(path, i)
                if i < max_items:
                    summary_str += format_item(i, item, item_path, d, s)
                else:
                    summary_str += (
                        f"{s*d}... {len(collection) - max_items} more items ...\n"
                    )
                    break
            return summary_str

        summary_str = ""
        for ix, key in enumerate(self.keys()):
            if ix >= max_items:
                summary_str += (
                    f"{sep*depth} ... {len(self.keys()) - max_items} more keys ...\n"
                )
                break
            new_path = format_path(current_path, key)
            summary_str += format_item(key, self[key], new_path, depth, sep)
        return summary_str

    def print_summary(self, **kwargs):
        from builtins import print

        print(self.summary(**kwargs))

    def write_summary(self, to, **kwargs):
        writelines(self.summary(sep="\t", **kwargs).split("\n"), to, mode="w")

    write = write_summary

    def fetch(self, addr):
        if isinstance(addr, (list, L)):
            return L([self.fetch(_addr) for _addr in addr])

        o = self
        for p in addr.split("."):
            try:
                o = o[int(p)]
            except:
                o = o[p]
        return o

    def merge_addersses(self, ADs):
        """Given a list of ADs, merge them into a single AD, which recursively merging the addresses
        E.g. When input is
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc.metrics.old.0.recall - 0.5 (🏷️ float)
        ```
        ,
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc.metrics.new.0.recall - 0.7566584022717563 (🏷️ float)
        ```
        ,
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc.metrics.new.1.recall - 0.8080808080808081 (🏷️ float)
        ```
        ,
        ```↯ AttrDict ↯
        01f3db47-f33a-11ee-a30d-129eb4343ebc.metrics.old.0.recall - 0.5 (🏷️ float)
        ```
        Output will be
        ```↯ AttrDict ↯
        01f3daf0-f33a-11ee-a30b-129eb4343ebc
          metrics
            old[]
              0
                recall - 0.5 (🏷️ float)
            new[]
              0
                recall - 0.7566584022717563 (🏷️ float)
              1
                recall - 0.8080808080808081 (🏷️ float)
        01f3db47-f33a-11ee-a30d-129eb4343ebc
          metrics
            old[]
              0
                recall - 0.5 (🏷️ float)
        ```
        """
        o = AttrDict()
        for ad in ADs:
            for k in ad.keys():
                o.set(k, ad[k])
        return o

    def __or__(self, other):
        if isinstance(other, dict):
            other = AD(**other)
        return self.merge_addersses([self.flatten(), other.flatten()])

    def flatten(self):
        """Flatten the AD into a single level AD with keys as dot combined keys"""
        o = AttrDict()
        for k in self.keys():
            v = self[k]
            if is_attrdict_like(v):
                v = v.flatten()
                for _k in v.keys():
                    o[f"{k}.{_k}"] = v[_k]
            elif isinstance(v, (list, tuple, set, frozenset, L)):
                for i, _v in enumerate(v):
                    if is_attrdict_like(_v):
                        _v = _v.flatten()
                        for _k, __v in _v.items():
                            o[f"{k}.{i}.{_k}"] = __v
                    else:
                        o[f"{k}.{i}"] = _v
            else:
                o[k] = v
        return o

    def fetch2(self, *, key=None, addrs=None):
        """given dot notation address/addresses fetch the value while maintaining the original structure"""
        if key is not None:
            addrs = self.find_address(key)
        if isinstance(addrs, (list, L)):
            return self.merge_addersses(
                L([self.fetch2(addrs=_addr) for _addr in addrs])
            )

        o = self
        for p in addrs.split("."):
            try:
                o = o[int(p)]
            except:
                o = o[p]
        return AttrDict({addrs: o})

    def write_config(self, to):
        c = Config(self.dict())
        c.to_disk(to)
        return P(to)

    def slice(self, key):
        ks = self.find_address(key)
        vs = self.fetch(ks)
        return AttrDict(dict(zip(ks, vs)))

    def flatten_and_make_dataframe(self):
        import pandas as pd

        df = pd.DataFrame([(*k.split("."), v) for k, v in self.flatten().items()])
        return df

    def diff(self, other):
        return AD(DeepDiff(self.dict(), AD(other).dict()))

    @property
    def d(self):
        return self.dict()


AD = AttrDict
AD.to_dict = AD.dict


def pretty_json(
    i, fpath=None, indent=4, print_with_logger=True, return_as_string=False
):
    def set_default(obj):
        if isinstance(obj, (set, BB, L)):
            return list(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, P):
            return str(obj)

    assert isinstance(i, (dict, list))
    i = unpack(i)
    dump = json.dumps(i, indent=indent, default=set_default)
    if fpath:
        makedir(parent(fpath))
        logger.info(f"Dumped a pretty file to {fpath}")
        with open(fpath, "w") as f:
            json.dump(i, f, indent=indent, default=set_default)
            return
    if print_with_logger:
        return logger.opt(depth=1).log("DEBUG", f"\n{dump}")
    else:
        if return_as_string:
            return dump
        print(dump)


# %% ../nbs/markups.ipynb 8
def read_json(fpath):
    import json

    with open(fpath, "r") as f:
        return json.load(f)


def write_json(obj, fpath, silent=False):
    from datetime import date, datetime

    def set_default(obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError

    if not silent:
        logger.opt(depth=1).log("DEBUG", f"Dumping json to {fpath}")
    with open(fpath, "w") as f:
        json.dump(obj, f, indent=4, default=set_default)
    return P(fpath)


# %% ../nbs/markups.ipynb 10
def write_jsonl(items, dest, mode="a"):
    makedir(parent(dest))
    with jsonlines.open(dest, mode) as writer:
        writer.write_all(items)
        if mode == "a":
            Info(f"Appended {len(items)} items to {dest}")
        elif mode == "w":
            Info(f"Wrote {len(items)} jsons to {dest}")


def read_jsonl(file):
    return [json.loads(line) for line in readlines(file, silent=True)]


# %% ../nbs/markups.ipynb 11
def read_yaml(file):
    with open(file, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def write_yaml(content, fpath):
    with open(fpath, "w") as outfile:
        yaml.dump(content, outfile, default_flow_style=False)


# %% ../nbs/markups.ipynb 12
def read_xml(file_path: Union[str, P]) -> AttrDict:
    "Read xml data as a dictionary"
    with open(str(file_path)) as xml_file:
        data = AttrDict(given_input_to_ad=xmltodict.parse(xml_file.read()))
    return data


def write_xml(data: Union[AttrDict, dict], file_path: Union[str, P]):
    makedir(parent(file_path))
    "convert a dictionary to xml"
    with open(file_path, "w") as xml_file:
        data = data.dict() if isinstance(data, AttrDict) else data
        assert isinstance(data, dict), "Function only supports dicts for now"
        data = xmltodict.unparse(data, pretty=True)
        xml_file.write(data)


def decompose(i):
    print(
        AD(
            type=str(type(i)),
            **{k: getattr(i, k) for k in dir(i) if not k.startswith("_")},
        )
    )


def basic_ad_repr(flds=None):
    "Minimal `__repr__` inspired from fastcore"
    import re

    if isinstance(flds, str):
        flds = re.split(", *", flds)
    flds = list(flds or [])

    def _f(self):
        m = str(type(self).__module__) + "."
        if m == "__main__.":
            m = ""
        res = f"{m}{type(self).__name__}"
        fs = flds if flds else [o for o in vars(self) if not o.startswith("_")]
        sig = AD({o: getattr(self, o) for o in fs}).summary(depth=1)
        return f"{res}(\n{sig})"

    return _f


# %% ../nbs/markups.ipynb 13
Config = Config

if __name__ == "__main__":
    assert AD({}) == {}

    p = 1
    q = {"a": 10}
    s = AD(p, q, a=20, b=30)
    assert s == {"a": 20, "b": 30, "p": 1, "q": {"a": 10}}

    s = AD(p, **q, b=30)
    assert s == {"a": 10, "b": 30, "p": 1}

    p = {"b": 222}
    assert AD(p, {"a": 2, "b": 2}, {"bb": 2, "a": 3, "b": 3}, a=20, b=30) == {
        "a": 2,
        "b": 2,
        "bb": 2,
        "p": {"b": 222},
    }

    d1 = {"c": 1}
    d2 = {"c": 2}
    d3 = {"c": 3}
    l = [1, 2, 3]
    assert AD(l, {"c": 100}, d1, d2, d3, d1={"c": 10}, d2={"c": 20}).dict() == {
        "l": [1, 2, 3],
        "d1": {"c": 1},
        "d2": {"c": 2},
        "c": 100,
        "d3": {"c": 3},
    }

    a = 20
    b = 30
    assert AD(a, b).dict() == {"a": 20, "b": 30}

    p = 10
    q = {"a": 1}
    assert AD(p, q) == {"p": 10, "q": {"a": 1}}
    assert AD(p, {"a": 1}) == {"p": 10, "a": 1}

    x = {"a": 1, "b": 2, "c": 3}
    assert AD(x).dict() == {"x": {"a": 1, "b": 2, "c": 3}}
    assert AD(**x).dict() == {"a": 1, "b": 2, "c": 3}
